/*
 * copyright (c)2009 http://wamsoft.jp
 * zlib license
 */

#ifndef __SQOBJECTINFO_H__
#define __SQOBJECTINFO_H__

#include <squirrel.h>

namespace sqobject {

/// 
HSQUIRRELVM init();
/// I
extern void done();
/// ێpO[oVM̎擾
extern HSQUIRRELVM getGlobalVM();

class Object;
class Thread;
	
// ---------------------------------------------------------
// ObjectInfo
// ---------------------------------------------------------

/**
 * squirrel IuWFNgێpNX
 * QƂێ\
 */
class ObjectInfo {

public:
	/// r
	bool operator ==(const ObjectInfo& o);

	// e
	void clear();

	// X^bN擾
	void getStack(HSQUIRRELVM v, int idx);

    // X^bNQƂƂĎ擾
	void getStackWeak(HSQUIRRELVM v, int idx);
  
	// RXgN^
	ObjectInfo();

	// RXgN^
	ObjectInfo(HSQUIRRELVM v, int idx);

	// Rs[RXgN^
	ObjectInfo(const ObjectInfo &orig);

	// 
	ObjectInfo & operator=(const ObjectInfo &orig);

	// fXgN^
	virtual ~ObjectInfo();

	// null H
	bool isNull() const;
	
	// XbhH
	bool isSameThread(const HSQUIRRELVM v) const;

	// Xbh擾
	operator HSQUIRRELVM() const;
	
	// IuWFNgPUSH
	void push(HSQUIRRELVM v) const;

	/// o^(sNULL)
	void pushClone(HSQUIRRELVM v) const;
	
	// ---------------------------------------------------
	// delegate p
	// ---------------------------------------------------

	// delegate Ƃċ@\邩ǂ
	bool isDelegate() const;

	// bindenv 邩ǂ
	bool isBindDelegate() const;

	// ---------------------------------------------------
	// f[^擾
	// ---------------------------------------------------

	const SQChar *getString();
	
	// ---------------------------------------------------
	// IuWFNg擾
	// ---------------------------------------------------

	// Thread ̃CX^X[U|C^擾
	Thread *getThread();
	
	// Object ̃CX^X[U|C^擾
	Object *getObject();
	
	// ---------------------------------------------------
	// waitp\bh
	// ---------------------------------------------------

	bool isSameString(const SQChar *str) const;

	// ---------------------------------------------------
	// zEp
	// ---------------------------------------------------

	/// zƂď
	void initArray(int size=0);
	
	/// @return zȂ true
	bool isArray() const { return sq_isarray(obj); }

	/// zƂď
	void initTable();

	/// @return zȂ true
	bool isTable() const { return sq_istable(obj); }
	
	/// zɒlǉ
	void append(HSQUIRRELVM v, int idx);

	/// zɔzǉ
	void appendArray(ObjectInfo &array);
	
	/// zɒlǉ
	template<typename T>
	void append(T value) {
		HSQUIRRELVM gv = getGlobalVM();
		push(gv);
		pushValue(gv, value);
		sq_arrayappend(gv, -2);
		sq_pop(gv,1);
	}

	/// zɒl}
	template<typename T>
	void insert(int index, T value) {
		HSQUIRRELVM gv = getGlobalVM();
		push(gv);
		pushValue(gv, value);
		sq_arrayinsert(gv, -2, index);
		sq_pop(gv,1);
	}

	/// z̒l폜
	void remove(int index) {
		HSQUIRRELVM gv = getGlobalVM();
		push(gv);
		sq_arrayremove(gv, -1, index);
		sq_pop(gv,1);
	}
	
	/// z/ɒli[
	template<typename K, typename T>
	void set(K key, T value) {
		HSQUIRRELVM gv = getGlobalVM();
		push(gv);
		pushValue(gv, key);
		pushValue(gv, value);
		sq_set(gv, -3);
		sq_pop(gv,1);
	}

	/// ɒlVKi[
	template<typename K, typename T>
	void create(K key, T value) {
		HSQUIRRELVM gv = getGlobalVM();
		push(gv);
		pushValue(gv, key);
		pushValue(gv, value);
		sq_createslot(gv, -3);
		sq_pop(gv,1);
	}
	
	/// z/̒l擾
	template<typename K>
	ObjectInfo get(K key) {
		HSQUIRRELVM gv = getGlobalVM();
		pushData(gv, key);
		ObjectInfo ret(gv, -1);
		sq_pop(gv,1);
		return ret;
	}

	/// z/̒l擾B݂Ȃnull
	template<typename K, typename T>
	SQRESULT get(K key, T **value) {
		HSQUIRRELVM gv = getGlobalVM();
		pushData(gv, key);
		SQRESULT ret = getValue(gv, value);
		sq_pop(gv,1);
		return ret;
	}
	
	/// z̒lpushBȂnull
	template<typename T>
	void pushData(HSQUIRRELVM v, T key) const {
		push(v);
		pushValue(v, key);
		if (SQ_FAILED(sq_get(v, -2))) {
			sq_pushnull(v);
		}
		sq_remove(v,-2);
	}
	
	/// zE̒gNA
	void clearData() {
		HSQUIRRELVM gv = getGlobalVM();
		push(gv);
		sq_clear(gv,-1);
		sq_pop(gv,1);
	}

	/// @return z̒
	int len() const;

	/**
	 * z̓eSPUSH
	 * @param v squirrelVM
	 * @return push 
	 */
	int pushArray(HSQUIRRELVM v) const;

	// ---------------------------------------------------
	// ֐p
	// ---------------------------------------------------

	/// @return zȂ true
	bool isClosure() const { return sq_isclosure(obj) || sq_isnativeclosure(obj); }

	// Ăяo
	SQRESULT call();

	template<typename T1> SQRESULT call(T1 p1) {
		HSQUIRRELVM gv = getGlobalVM();
		push(gv);
		sq_pushroottable(gv); // root
		pushValue(gv, p1);
		SQRESULT ret = sq_call(gv, 2, SQFalse, SQTrue);
		sq_pop(gv, 1);
		return ret;
	}

	template<typename T1, typename T2> SQRESULT call(T1 p1, T2 p2) {
		HSQUIRRELVM gv = getGlobalVM();
		push(gv);
		sq_pushroottable(gv); // root
		pushValue(gv, p1);
		pushValue(gv, p2);
		SQRESULT ret = sq_call(gv, 3, SQFalse, SQTrue);
		sq_pop(gv, 1);
		return ret;
	}
	
	template<typename R> SQRESULT call(R* r) {
		SQRESULT ret;
		HSQUIRRELVM gv = getGlobalVM();
		push(gv);
		sq_pushroottable(gv); // root
		if (SQ_SUCCEEDED(ret = sq_call(gv, 1, SQTrue, SQTrue))) {
			ret = getValue(gv, r);
			sq_pop(gv, 1);
		}
		sq_pop(gv, 1);
		return ret;
	}
	template<typename R, typename T1> SQRESULT call(R* r, T1 p1) {
		SQRESULT ret;
		HSQUIRRELVM gv = getGlobalVM();
		push(gv);
		sq_pushroottable(gv); // root
		pushValue(gv, p1);
		if (SQ_SUCCEEDED(ret = sq_call(gv, 2, SQTrue, SQTrue))) {
			ret = getValue(gv, r);
			sq_pop(gv, 1);
		}
		sq_pop(gv, 1);
		return ret;
	}
	template<typename R, typename T1, typename T2> SQRESULT call(R* r, T1 p1, T2 p2) {
		SQRESULT ret;
		HSQUIRRELVM gv = getGlobalVM();
		push(gv);
		sq_pushroottable(gv); // root
		pushValue(gv, p1);
		pushValue(gv, p2);
		if (SQ_SUCCEEDED(ret = sq_call(gv, 3, SQTrue, SQTrue))) {
			ret = getValue(gv, r);
			sq_pop(gv, 1);
		}
		sq_pop(gv, 1);
		return ret;
	}
	
protected:
	// ---------------------------------------------------
	// X^bNp
	// ---------------------------------------------------

	// l push
	static void pushValue(HSQUIRRELVM v) {}
	static void pushValue(HSQUIRRELVM v, int value) { sq_pushinteger(v,value); }
	static void pushValue(HSQUIRRELVM v, const SQChar *value) { sq_pushstring(v,value,-1); }
	static void pushValue(HSQUIRRELVM v, SQUserPointer value) { sq_pushuserpointer(v,value); }
	static void pushValue(HSQUIRRELVM v, ObjectInfo &obj) { obj.push(v); }
	
	// l̎擾
	static SQRESULT getValue(HSQUIRRELVM v, void *value) { return SQ_OK; }
	static SQRESULT getValue(HSQUIRRELVM v, int *value) { return sq_getinteger(v, -1, value); }
	static SQRESULT getValue(HSQUIRRELVM v, const SQChar **value) { return sq_getstring(v, -1, value); }
	static SQRESULT getValue(HSQUIRRELVM v, SQUserPointer *value) { return sq_getuserpointer(v, -1, value); }
	
private:
	HSQOBJECT obj; // IuWFNgQƏ

	// -------------------------------------------------------------
	// Xbhnp
	// -------------------------------------------------------------

public:
	// Object
	void addWait(ObjectInfo &thread);
	void removeWait(ObjectInfo &thread);

	// Thread
	bool notifyTrigger(const SQChar *name);
	bool notifyObject(ObjectInfo &target);
};

};

#endif
